/*
 * Copyright (c) 2004-2007 The Trustees of Indiana University and Indiana
 *                         University Research and Technology
 *                         Corporation.  All rights reserved.
 * Copyright (c) 2004-2008 The University of Tennessee and The University
 *                         of Tennessee Research Foundation.  All rights
 *                         reserved.
 * Copyright (c) 2004-2005 High Performance Computing Center Stuttgart, 
 *                         University of Stuttgart.  All rights reserved.
 * Copyright (c) 2004-2005 The Regents of the University of California.
 *                         All rights reserved.
 * $COPYRIGHT$
 * 
 * Additional copyrights may follow
 * 
 * $HEADER$
 */

#include "ompi_config.h"

#include <string.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/mman.h>
#include <unistd.h>
#include <stddef.h>
#include "opal/class/opal_bitmap.h"
#include "opal/datatype/opal_convertor.h"
#include "opal/sys/atomic.h"
#include "ompi/mca/btl/btl.h"
#include "ompi/mca/mpool/base/base.h"
#include "btl_hlink.h"
#include "btl_hlink_frag.h"
#include "ompi/proc/proc.h"

extern void btl_hlink_dma_copy (mpm_transport_h *h, uint32_t dest, uint32_t source, uint32_t buff_size);
extern void btl_hlink_dma_linked_copy (mpm_transport_h *h, int num_links, uint32_t dest[], uint32_t source[], uint32_t buff_size[]);
/***************************************************************************************************/
mca_btl_hlink_t mca_btl_hlink = {
 {
    &mca_btl_hlink_component.super,
    0, /* btl_eager_limit */
    0, /* btl_rndv_eager_limit */
    0, /* btl_max_send_size */
    0, /* btl_rdma_pipeline_send_length */
    0, /* btl_rdma_pipeline_frag_size */
    0, /* btl_min_rdma_pipeline_size */
    0, /* btl_exclusivity */
    0, /* btl_latency */
    0, /* btl_bandwidth */
    0, /* btl flags */
    0, /* btl segment size */
    mca_btl_hlink_add_procs,
    mca_btl_hlink_del_procs,
    NULL,
    mca_btl_hlink_finalize,
    mca_btl_hlink_alloc,
    mca_btl_hlink_free,
    mca_btl_hlink_prepare_src,
    mca_btl_hlink_prepare_dst,
    mca_btl_hlink_send,
    mca_btl_hlink_sendi,
    NULL,
    NULL,
    mca_btl_base_dump,
    NULL, /* mpool */
    NULL, /* register error cb */
    mca_btl_hlink_ft_event
  }
};
/***************************************************************************************************/
int mca_btl_hlink_find_rmt_ep (int endpoint_count, int hlink_iface, int my_rank, int peer_rank) 
{
int rmt_ep = 0;

    goto END_OF_FIND_RMT_EP; //At the moment we are limitting to one MPI process per SOC

    if(endpoint_count == 4)
    {
      if       (my_rank == 0) {
        if(peer_rank == 0) rmt_ep = 0;
        else               rmt_ep = 2;
      } else if(my_rank == 1) {
        if(peer_rank == 0) rmt_ep = 1;
        else               rmt_ep = 3;
      } else if(my_rank == 2) {
        if(peer_rank == 0) rmt_ep = 0;
        else               rmt_ep = 2;
      } else if(my_rank == 3) {
        if(peer_rank == 0) rmt_ep = 1;
        else               rmt_ep = 3;
      } //error condition
    }
END_OF_FIND_RMT_EP:
    return rmt_ep;
}
/***************************************************************************************************/
int
mca_btl_hlink_module_init (mca_btl_hlink_t *hlink_module,
                          void *dev)
{
    int rc;

    BTL_VERBOSE(("binding module %p to device %p", (void *) hlink_module,
                 (void *) dev));

    /* copy module defaults (and function pointers) */
    memmove (hlink_module, &mca_btl_hlink, sizeof (mca_btl_hlink_t));

    hlink_module->endpoints = NULL;
    mca_btl_hlink_component.endpoint_count = 0;
//    dev->btl_ctx = (void *) hlink_module;

    return OMPI_SUCCESS;
}
/***************************************************************************************************/
int mca_btl_hlink_init_ep (int local_proc, int hlink_iface, int bridging,
                           mca_btl_base_endpoint_t **ep,
                           mca_btl_hlink_t *btl,
                           ompi_proc_t *peer_proc) 
{
    mca_btl_base_endpoint_t *endpoint;
    int rc;

    endpoint = calloc(1, sizeof(mca_btl_base_endpoint_t));
    assert (endpoint != NULL);

    endpoint->btl = (void *)btl;
    endpoint->peer_rank = local_proc;
    endpoint->hlink_iface = hlink_iface;
    endpoint->bridging = bridging;
    *ep = endpoint;

    return OMPI_SUCCESS;
}

/***************************************************************************************************/
int mca_btl_hlink_add_procs( struct mca_btl_base_module_t* btl, 
                            size_t nprocs, 
                            struct ompi_proc_t **procs, 
                            struct mca_btl_base_endpoint_t **peers,
                            opal_bitmap_t* reachability )
{
    int i;
    size_t ntotal_procs, nlocal_procs;
    mca_btl_hlink_t *hlink_module = (mca_btl_hlink_t *) btl;
    bool first_time_init = (NULL == hlink_module->endpoints);
    ompi_proc_t* my_proc; /* pointer to caller's proc structure */
    int rc;

    /* get pointer to my proc structure */
    if(NULL == (my_proc = ompi_proc_local())) return OMPI_ERR_OUT_OF_RESOURCE;

    if (NULL == hlink_module->endpoints) 
    {
      (void) ompi_proc_world (&ntotal_procs);

      hlink_module->endpoints = calloc (ntotal_procs, sizeof (mca_btl_base_endpoint_t *));
      if (OPAL_UNLIKELY(NULL == hlink_module->endpoints)) return OMPI_ERR_OUT_OF_RESOURCE;

      /* initialize own FIFOs */
      /*
       * The receiver initializes all its FIFOs.  All components will
       * be allocated near the receiver.  Nothing will be local to
       * "the sender" since there will be many senders.
       */
      //if(nprocs > BTL_HLINK_MAX_PROCESS) return OMPI_ERR_OUT_OF_RESOURCE;

      if(mca_btl_hlink_component.use_hl0)
      {
        //for(int i = 0; i < nprocs; i++)
        for(int i = 0; i < BTL_HLINK_MAX_LOCAL_PROCESS; i++)
        { // Local processes that can communicate with the remote node
          mca_btl_hlink_component.fifo[0][i] =
            (hlink_fifo_t *)((char *)mca_btl_hlink_component.hl_win_msmc_loc + i * btl_hlink_align64(hlink_fifo_getsize(BTL_HLINK_MAX_FIFO_ITEMS)));
          mca_btl_hlink_component.fifo[1][i] =
            (hlink_fifo_t *)((char *)mca_btl_hlink_component.hl_win_msmc_rmt + i * btl_hlink_align64(hlink_fifo_getsize(BTL_HLINK_MAX_FIFO_ITEMS)));
          //Initilize fifo-s on local side only:
          hlink_fifo_init(mca_btl_hlink_component.fifo[0][i], BTL_HLINK_MAX_FIFO_ITEMS);

          mca_btl_hlink_component.hl_win_rmt_msg_idx[i] = 0;
          mca_btl_hlink_component.hl_win_rmt_msgbase[i] =
           (void *)((char *)mca_btl_hlink_component.hl_win_rmt + i * BTL_HLINK_PAYLOAD_SIZE);
          mca_btl_hlink_component.hl_win_loc_msgbase[i] =
           (void *)((char *)mca_btl_hlink_component.hl_win_loc + i * BTL_HLINK_PAYLOAD_SIZE);
          opal_output_verbose(100, mca_btl_base_output, "TIMPIDBG %d rmt_msgbase=%p loc_msgbase=%p", i, 
                       mca_btl_hlink_component.hl_win_rmt_msgbase[i], mca_btl_hlink_component.hl_win_loc_msgbase[i]);
        }
      }
      if(mca_btl_hlink_component.use_hl1)
      {
        //for(int i = 0; i < nprocs; i++)
        for(int i = 0; i < BTL_HLINK_MAX_LOCAL_PROCESS; i++)
        { // Local processes that can communicate with the remote node
          mca_btl_hlink_component.fifo1[0][i] =
            (hlink_fifo_t *)((char *)mca_btl_hlink_component.hl_win_msmc_loc1 + i * btl_hlink_align64(hlink_fifo_getsize(BTL_HLINK_MAX_FIFO_ITEMS)));
          mca_btl_hlink_component.fifo1[1][i] =
            (hlink_fifo_t *)((char *)mca_btl_hlink_component.hl_win_msmc_rmt1 + i * btl_hlink_align64(hlink_fifo_getsize(BTL_HLINK_MAX_FIFO_ITEMS)));
          //Initilize fifo-s on local side only:
          hlink_fifo_init(mca_btl_hlink_component.fifo1[0][i], BTL_HLINK_MAX_FIFO_ITEMS);

          mca_btl_hlink_component.hl_win_rmt1_msg_idx[i] = 0;
          mca_btl_hlink_component.hl_win_rmt1_msgbase[i] =
           (void *)((char *)mca_btl_hlink_component.hl_win_rmt1 + i * BTL_HLINK_PAYLOAD_SIZE);
          mca_btl_hlink_component.hl_win_loc1_msgbase[i] =
           (void *)((char *)mca_btl_hlink_component.hl_win_loc1 + i * BTL_HLINK_PAYLOAD_SIZE);
          opal_output_verbose(100, mca_btl_base_output, "TIMPIDBG %d rmt_msgbase=%p loc_msgbase=%p", i,
                       mca_btl_hlink_component.hl_win_rmt1_msgbase[i], mca_btl_hlink_component.hl_win_loc1_msgbase[i]);
        }
      }
      //Diagonal
      if((nprocs > 2) && 
          mca_btl_hlink_component.hl_win_msmc_rmt_rd_bridge && 
          mca_btl_hlink_component.hl_win_msmc_rmt_wr_bridge)
      {
          for(int i = 0; i < BTL_HLINK_MAX_LOCAL_PROCESS; i++)
          { // Local processes that can communicate with the remote node
            mca_btl_hlink_component.fifo_bridge[0][i] =
              (hlink_fifo_t *)((char *)mca_btl_hlink_component.hl_win_msmc_rmt_rd_bridge + i * btl_hlink_align64(hlink_fifo_getsize(BTL_HLINK_MAX_FIFO_ITEMS)));
            mca_btl_hlink_component.fifo_bridge[1][i] =
              (hlink_fifo_t *)((char *)mca_btl_hlink_component.hl_win_msmc_rmt_wr_bridge + i * btl_hlink_align64(hlink_fifo_getsize(BTL_HLINK_MAX_FIFO_ITEMS)));
            //Initilize fifo-s on local side only:
            hlink_fifo_init(mca_btl_hlink_component.fifo_bridge[0][i], BTL_HLINK_MAX_FIFO_ITEMS);

            mca_btl_hlink_component.hl_win_rmt_wr_bridge_msg_idx[i] = 0;
            mca_btl_hlink_component.hl_win_rmt_wr_bridge_msgbase[i] =
             (void *)((char *)mca_btl_hlink_component.hl_win_rmt_wr_bridge + i * BTL_HLINK_PAYLOAD_SIZE);
            mca_btl_hlink_component.hl_win_rmt_rd_bridge_msgbase[i] =
             (void *)((char *)mca_btl_hlink_component.hl_win_rmt_rd_bridge + i * BTL_HLINK_PAYLOAD_SIZE);
            opal_output_verbose(100, mca_btl_base_output, "TIMPIDBG BRIDGE %d rmt_wr_bridge_msgbase=%p rmt_rd_bridge_msgbase=%p", i,
                         mca_btl_hlink_component.hl_win_rmt_wr_bridge_msgbase[i], mca_btl_hlink_component.hl_win_rmt_rd_bridge_msgbase[i]);
          }
          if(mca_btl_hlink_component.hl_win_rmt_rd_bridge_hostname)
          {
             char bridge_hostname[80];
             memset (bridge_hostname, 0, 80);
             memcpy (bridge_hostname, mca_btl_hlink_component.hl_win_rmt_rd_bridge_hostname, 78);
             opal_output_verbose(20, mca_btl_base_output, "diagonal peer is:[%s]", bridge_hostname);
          } else {
             opal_output_verbose(20, mca_btl_base_output, " diagonal peer NOT ACCEESIBLE[%s]");
          } 
      }

    }
    for (i = 0, nlocal_procs = 0 ; i < nprocs ; ++i) 
    {
        struct ompi_proc_t *ompi_proc = procs[i];
        uint32_t rem_rank = ompi_proc->proc_name.vpid;
        uint32_t rem_jobid = ompi_proc->proc_name.jobid;
        char *rmt_hl_hostname = (char *)mca_btl_hlink_component.hl_win_rmt_hostname;
        char *rmt1_hl_hostname = (char *)mca_btl_hlink_component.hl_win_rmt1_hostname;
        char *rmt_bridge_hl_hostname = (char *)mca_btl_hlink_component.hl_win_rmt_rd_bridge_hostname;
        int hlink_iface = 0;
        int bridging    = 0;
        opal_output_verbose(100, mca_btl_base_output, "TIMPIDBG B hlink_module_add_procs(%d) remote hostname=%s (over HL0:%s HL1:%s, bridge:%s) rem_rank[%d] jobid=%08x\n", 
        i, ompi_proc->proc_hostname, 
        mca_btl_hlink_component.use_hl0 ? rmt_hl_hostname : "NOTUSED", 
        mca_btl_hlink_component.use_hl1 ? rmt1_hl_hostname : "NOTUSED",
        rmt_bridge_hl_hostname ? rmt_bridge_hl_hostname : "NOTUSED", 
        rem_rank, rem_jobid);
        /* check to see if this is me */
        if(my_proc == procs[i]) {
           mca_btl_hlink_component.my_rank = i;
        }

        if (OPAL_PROC_ON_LOCAL_NODE(ompi_proc->proc_flags)) {
            /* don't use HLINK to communicate with local procs */
            continue;
        } 
 
        if(mca_btl_hlink_component.use_hl0 && (strncmp(ompi_proc->proc_hostname, rmt_hl_hostname, 80) == 0))
        {
           hlink_iface = 0;
           opal_output_verbose(100, mca_btl_base_output, "HLINK0: TIMPIDBG[%d, nlocal %d] hlink_module_add_procs, I can talk to rank: (%d) on %s", i, nlocal_procs,
                        rem_rank, rmt_hl_hostname);
           nlocal_procs++;
        } else if(mca_btl_hlink_component.use_hl1 && (strncmp(ompi_proc->proc_hostname, rmt1_hl_hostname, 80) == 0))
        {
          hlink_iface = 1;
          opal_output_verbose(100, mca_btl_base_output, "HLINK1: TIMPIDBG[%d, nlocal %d] hlink_module_add_procs, I can talk to rank: (%d) on %s", i, nlocal_procs,
                          rem_rank, rmt1_hl_hostname);
          nlocal_procs++;
        } else if((nprocs > 2) && mca_btl_hlink_component.hl_win_rmt_rd_bridge_hostname && 
                  strncmp(ompi_proc->proc_hostname, mca_btl_hlink_component.hl_win_rmt_rd_bridge_hostname, 80) == 0)
        {  //At least one interface is enabled
           hlink_iface = mca_btl_hlink_component.node_id == '1' ? 0 : 1;
           bridging    = 1;
           opal_output_verbose(100, mca_btl_base_output, "HLINK%dbridge: TIMPIDBG[%d, nlocal %d] hlink_module_add_procs, I can talk to rank: (%d) on %s", 
                        hlink_iface, i, nlocal_procs, rem_rank, rmt_bridge_hl_hostname);
           nlocal_procs++;
        } else {
           opal_output_verbose(100, mca_btl_base_output, "TIMPIDBG[%d] hlink_module_add_procs, I can NOT talk to rank: (%d)", i, rem_rank);
           continue;
        }

        /*  Create and Init endpoints */
        rc = mca_btl_hlink_init_ep (nlocal_procs - 1, hlink_iface, bridging, peers + i, (mca_btl_hlink_t *) btl, ompi_proc);

        if (OPAL_UNLIKELY(OMPI_SUCCESS != rc)) {
            BTL_ERROR(("btl/hlink error initializing endpoint"));
            return rc;
        }
        /* Set the reachable bit */
        rc = opal_bitmap_set_bit (reachability, i);
        /* Store a reference to this peer */
        hlink_module->endpoints[rem_rank] = peers[i];
    }
    mca_btl_hlink_component.hlink_nprocs = nlocal_procs;
    /******************************************************************************************/
    mca_btl_hlink_component.endpoint_count += nprocs;
    opal_output_verbose(100, mca_btl_base_output, "TIMPIDBG hlink_module_add_procs, RANKS: my_rank=%d nprocs=%d nlocal_procs=%d endpoint_count=%d\n", 
       mca_btl_hlink_component.my_rank, nprocs, nlocal_procs, mca_btl_hlink_component.endpoint_count);
    return OMPI_SUCCESS;
}
/***************************************************************************************************/

int mca_btl_hlink_del_procs( struct mca_btl_base_module_t* btl, 
                            size_t nprocs,
                            struct ompi_proc_t **procs, 
                            struct mca_btl_base_endpoint_t **peers )
{
    opal_output_verbose(20, mca_btl_base_output, "TIMPIDBG: hlink_module_del_procs!!!");
    return OMPI_SUCCESS;
}


/******************************************************************************/
/**
 * MCA->BTL Clean up any resources held by BTL module
 * before the module is unloaded.
 *
 * @param btl (IN)   BTL module.
 *
 * Prior to unloading a BTL module, the MCA framework will call
 * the BTL finalize method of the module. Any resources held by
 * the BTL should be released and if required the memory corresponding
 * to the BTL module freed.
 *
 */

int mca_btl_hlink_finalize(struct mca_btl_base_module_t* btl)
{
  opal_output_verbose(20, mca_btl_base_output, "hlink_module_finalize!!!");
  sleep(1); /* Let all in-flight transfers end */

  if(mca_btl_hlink_component.hl_win_rmt_rd_bridge_hostname) memset (mca_btl_hlink_component.hl_win_rmt_rd_bridge_hostname, 0, 80);
  if(mca_btl_hlink_component.use_hl0) {
    char *ptr_hostname = (char *)mca_btl_hlink_component.hl_win_msmc_loc + BTL_HLINK_MSG_OFFSET - 80;
    memset (ptr_hostname, 0, 80);
    mpm_transport_close (mca_btl_hlink_component.mpm_transport_hyplnk0);
  }
  if(mca_btl_hlink_component.use_hl1) {
    char *ptr_hostname = (char *)mca_btl_hlink_component.hl_win_msmc_loc1 + BTL_HLINK_MSG_OFFSET - 80;
    memset (ptr_hostname, 0, 80);
    mpm_transport_close (mca_btl_hlink_component.mpm_transport_hyplnk1);
  }
  mpm_transport_close (mca_btl_hlink_component.mpm_transport_shmem);
  return OMPI_SUCCESS;
}

/******************************************************************************/

/**
 * Allocate a segment.
 *
 * @param btl (IN)      BTL module
 * @param size (IN)     Request segment size.
 */
mca_btl_base_descriptor_t* mca_btl_hlink_alloc(
        struct mca_btl_base_module_t* btl,
        struct mca_btl_base_endpoint_t* endpoint,
        uint8_t order,
        size_t size,
        uint32_t flags)
{
    mca_btl_hlink_frag_t* frag = NULL;
    int rc;

    opal_output_verbose(100, mca_btl_base_output, "TIMPIDBG: hlink_module_alloc (size=%d)!!!", size);
    if(size <= mca_btl_hlink.super.btl_eager_limit) {
        MCA_BTL_HLINK_FRAG_ALLOC_EAGER(frag,rc);
    } else if (size <= mca_btl_hlink_component.max_frag_size) {
        MCA_BTL_HLINK_FRAG_ALLOC_MAX(frag,rc);
    }

    if (OPAL_LIKELY(frag != NULL)) {
        frag->segment.seg_len = size;
        frag->base.des_flags = flags;
    }
    return (mca_btl_base_descriptor_t*)frag;
}

/******************************************************************************/
/**
 * Return a segment allocated by this BTL.
 *
 * @param btl (IN)      BTL module
 * @param segment (IN)  Allocated segment.
 */
extern int mca_btl_hlink_free(
    struct mca_btl_base_module_t* btl,
    mca_btl_base_descriptor_t* des)
{
    mca_btl_hlink_frag_t* frag = (mca_btl_hlink_frag_t*)des;

    opal_output_verbose(100, mca_btl_base_output, "TIMPIDBG: hlink_module_free!!!");
    MCA_BTL_HLINK_FRAG_RETURN(frag);

    return OMPI_SUCCESS;
}
/******************************************************************************/


/**
 * Prepare data for send/put
 *
 * @param btl (IN)      BTL module
 */
struct mca_btl_base_descriptor_t*
mca_btl_hlink_prepare_src( struct mca_btl_base_module_t* btl,
                          struct mca_btl_base_endpoint_t* endpoint,
                          mca_mpool_base_registration_t* registration,
                          struct opal_convertor_t* convertor,
                          uint8_t order,
                          size_t reserve,
                          size_t* size,
                          uint32_t flags )
{
    mca_btl_hlink_frag_t* frag;
    struct iovec iov;
    uint32_t iov_count = 1;
    size_t max_data = *size;
    int rc;
    unsigned int *data_ptr;
    int jj;    

    opal_convertor_get_current_pointer (convertor, &data_ptr);

    opal_output_verbose(20, mca_btl_base_output, "TIMPIDBG: hlink_module_prepare_src EAGER: %d MAX: %d (btl=%p)",  
                        mca_btl_hlink_component.eager_limit, mca_btl_hlink_component.max_frag_size, btl);

        if ( reserve + max_data <= mca_btl_hlink_component.eager_limit ) {
            MCA_BTL_HLINK_FRAG_ALLOC_EAGER(frag,rc);
        } else {
            MCA_BTL_HLINK_FRAG_ALLOC_MAX(frag, rc);
        }
        if( OPAL_UNLIKELY(NULL == frag) ) {
            return NULL;
        }

        if( OPAL_UNLIKELY(reserve + max_data > frag->size) ) {
            max_data = frag->size - reserve;
        }
        iov.iov_len = max_data;
        iov.iov_base =
            (IOVBASE_TYPE*)(((unsigned char*)(frag->segment.seg_addr.pval)) + reserve);

        rc = opal_convertor_pack(convertor, &iov, &iov_count, &max_data );
        if( OPAL_UNLIKELY(rc < 0) ) {
            MCA_BTL_HLINK_FRAG_RETURN(frag);
            return NULL;
        }
        frag->segment.seg_len = reserve + max_data;

    frag->base.des_src = &(frag->segment);
    frag->base.des_src_cnt = 1;
    frag->base.order = MCA_BTL_NO_ORDER;
    frag->base.des_dst = NULL;
    frag->base.des_dst_cnt = 0;
    frag->base.des_flags = flags;
    *size = max_data;
    return &frag->base;
}

/**
 * Prepare data for receive.
 */
struct mca_btl_base_descriptor_t*
mca_btl_hlink_prepare_dst( struct mca_btl_base_module_t* btl,
                          struct mca_btl_base_endpoint_t* endpoint,
                          mca_mpool_base_registration_t* registration,
                          struct opal_convertor_t* convertor,
                          uint8_t order,
                          size_t reserve,
                          size_t* size,
                          uint32_t flags )
{
    int rc;
    void *ptr;
    mca_btl_hlink_frag_t* frag;

    opal_output_verbose(20, mca_btl_base_output, "TIMPIDBG: hlink_module_prepare_dst!!!");
    MCA_BTL_HLINK_FRAG_ALLOC_USER(frag, rc);
    if(OPAL_UNLIKELY(NULL == frag)) {
        return NULL;
    }

    frag->segment.seg_len = *size;
    opal_convertor_get_current_pointer( convertor, &ptr );
    frag->segment.seg_addr.lval = (uint64_t)(uintptr_t) ptr;
    
    frag->base.des_src = NULL;
    frag->base.des_src_cnt = 0;
    frag->base.des_dst = (mca_btl_base_segment_t*)&frag->segment;
    frag->base.des_dst_cnt = 1;
    frag->base.des_flags = flags;
    return &frag->base;
}
/******************************************************************************/
 
/**
 * Initiate a send to the peer.
 *
 * @param btl (IN)      BTL module
 * @param peer (IN)     BTL peer addressing
 */

int mca_btl_hlink_send( struct mca_btl_base_module_t* btl,
                       struct mca_btl_base_endpoint_t* endpoint,
                       struct mca_btl_base_descriptor_t* des,
                       mca_btl_base_tag_t tag )
{
    mca_btl_hlink_t *hlink_module = (mca_btl_hlink_t *) btl;
    mca_btl_hlink_frag_t *frag = (mca_btl_hlink_frag_t *)des;
    unsigned int *rmt_values, i;
    unsigned int seg_len_align;
    int rmt_ep;
    char host_name[80];
    mca_btl_hlink_hdr_t *hdr;
    char *ptr;
    unsigned int addr_RMT_PHY, addr_MSMC_fifo_PHY;
    int head_cnt;
    void *edma_hyplnk;
    void *tmp_fifo;
    int *tmp_rmt_msg_idx;

    opal_output_verbose(20, mca_btl_base_output, "TIMPIDBG: hlink_module_send frag=%p (base=%p size=%lld)!!!\n", frag,
                        frag->segment.seg_addr.pval, frag->segment.seg_len ); 
    /* available header space */
    frag->hdr->len = frag->segment.seg_len;
    /* type of message, pt-2-pt, one-sided, etc */
    frag->hdr->tag = tag;
    frag->endpoint = endpoint;
    opal_output_verbose(20, mca_btl_base_output, "TIMPIDBG: hlink_module_send FIFO_WRITE: nprocs %d(%d)->%d, frag_hdr=%p eager_limit=%d\n", 
            mca_btl_hlink_component.hlink_nprocs, mca_btl_hlink_component.my_rank, endpoint->peer_rank, frag->hdr, mca_btl_hlink_component.eager_limit );
    /* TODO: write to window and update fifo pointer */
    rmt_ep = mca_btl_hlink_find_rmt_ep ((int)mca_btl_hlink_component.endpoint_count, endpoint->hlink_iface, 
                                        (int)mca_btl_hlink_component.my_rank, (int)endpoint->peer_rank);

    if(endpoint->bridging == 1) {
      ptr = (char *)mca_btl_hlink_component.hl_win_rmt_wr_bridge_msgbase[rmt_ep] + mca_btl_hlink_component.hl_win_rmt_wr_bridge_msg_idx[rmt_ep];
      addr_RMT_PHY = mca_btl_hlink_component.ddr_rmt_bridge_addr + ((unsigned int)ptr - (unsigned int)mca_btl_hlink_component.hl_win_rmt_wr_bridge);
      tmp_fifo = mca_btl_hlink_component.fifo_bridge[1][rmt_ep];
      addr_MSMC_fifo_PHY = mca_btl_hlink_component.msmc_rmt_bridge_addr + ((unsigned int)tmp_fifo - (unsigned int)mca_btl_hlink_component.hl_win_msmc_rmt_wr_bridge);
      tmp_rmt_msg_idx = &mca_btl_hlink_component.hl_win_rmt_wr_bridge_msg_idx[rmt_ep];
      edma_hyplnk = endpoint->hlink_iface ? mca_btl_hlink_component.mpm_transport_hyplnk0 : mca_btl_hlink_component.mpm_transport_hyplnk1;
    } else {
      if(mca_btl_hlink_component.use_hl0 && (endpoint->hlink_iface == 0)) {
        ptr = (char *)mca_btl_hlink_component.hl_win_rmt_msgbase[rmt_ep] + mca_btl_hlink_component.hl_win_rmt_msg_idx[rmt_ep];
        addr_RMT_PHY = mca_btl_hlink_component.ddr_rmt_addr + ((unsigned int)ptr - (unsigned int)mca_btl_hlink_component.hl_win_rmt);
        tmp_fifo = mca_btl_hlink_component.fifo[1][rmt_ep];
        addr_MSMC_fifo_PHY = mca_btl_hlink_component.msmc_rmt_addr + ((unsigned int)tmp_fifo - (unsigned int)mca_btl_hlink_component.hl_win_msmc_rmt);
        tmp_rmt_msg_idx = &mca_btl_hlink_component.hl_win_rmt_msg_idx[rmt_ep];
        edma_hyplnk = mca_btl_hlink_component.mpm_transport_hyplnk0;
      } else if(mca_btl_hlink_component.use_hl1 && (endpoint->hlink_iface == 1)) {
        ptr = (char *)mca_btl_hlink_component.hl_win_rmt1_msgbase[rmt_ep] + mca_btl_hlink_component.hl_win_rmt1_msg_idx[rmt_ep];
        addr_RMT_PHY = mca_btl_hlink_component.ddr_rmt1_addr + ((unsigned int)ptr - (unsigned int)mca_btl_hlink_component.hl_win_rmt1);
        tmp_fifo = mca_btl_hlink_component.fifo1[1][rmt_ep];
        addr_MSMC_fifo_PHY = mca_btl_hlink_component.msmc_rmt1_addr + ((unsigned int)tmp_fifo - (unsigned int)mca_btl_hlink_component.hl_win_msmc_rmt1);
        tmp_rmt_msg_idx = &mca_btl_hlink_component.hl_win_rmt1_msg_idx[rmt_ep];
        edma_hyplnk = mca_btl_hlink_component.mpm_transport_hyplnk1;
      } else {
        opal_output(0, "BTL_HLINK endpoint invalid translation (hl0=%d hl1=%d hlink_iface=%d)\n", 
          mca_btl_hlink_component.use_hl0, mca_btl_hlink_component.use_hl1, endpoint->hlink_iface); 
        return OMPI_ERR_OUT_OF_RESOURCE;
      }
    }
    opal_output_verbose(20, mca_btl_base_output, "TMPIDBG: rmt_ep %d send (hyplnk_iface%d) (bridging %d) (use_edma %d seg_len %lld eager_limit %d)", rmt_ep,
        endpoint->hlink_iface, endpoint->bridging, mca_btl_hlink_component.use_edma, frag->segment.seg_len, mca_btl_hlink_component.eager_limit);

    if(mca_btl_hlink_component.use_edma && (frag->segment.seg_len >=  mca_btl_hlink_component.eager_limit))
    { // Big chunks of data first go to MSMC then to remote DDR over hyperlink using EDMA
      ptr = mca_btl_hlink_component.hl_win_edma_transfer; //user space address of MSMC transfer region
    }

    memcpy (ptr, frag->hdr, sizeof(mca_btl_hlink_hdr_t));
    hdr = (mca_btl_hlink_hdr_t *)ptr;
    ptr += sizeof(mca_btl_hlink_hdr_t);
    seg_len_align = btl_hlink_align32((unsigned int)frag->segment.seg_len);
    memcpy (ptr, frag->segment.seg_addr.pval, seg_len_align);

    if(mca_btl_hlink_component.use_edma && (frag->segment.seg_len >=  mca_btl_hlink_component.eager_limit))
    { /** EDMA transfer **/
      unsigned int *head_ptr, dst_addr[2], src_addr[2], length[2];    
      opal_output_verbose(20, mca_btl_base_output, "BTL_HLINK send EDMA_TRANSFER (to %08x from %08x, size=%d!\n",  
                  addr_RMT_PHY, mca_btl_hlink_component.msmc_edma_transfer_addr, btl_hlink_align64(sizeof(mca_btl_hlink_hdr_t) + seg_len_align));
      /* This is double linked transfer: push a fragment to remote DDR, then push a mail to remote MSMC */
      while((head_cnt = hlink_fifo_acquire(tmp_fifo, *tmp_rmt_msg_idx)) < 0)
      {  
         opal_output_verbose(20, mca_btl_base_output, "BTL_HLINK no room in FIFO (send)!");
      }

      head_ptr = (unsigned int *)((char *)mca_btl_hlink_component.hl_win_edma_transfer + BTL_HLINK_MSMC_EDMA_MMAP_SIZE - sizeof(uint32_t));
      dst_addr[0] = addr_RMT_PHY;
      src_addr[0] = mca_btl_hlink_component.msmc_edma_transfer_addr;
      length[0]   = btl_hlink_align64(sizeof(mca_btl_hlink_hdr_t) + seg_len_align);
      head_cnt    ++;
      *head_ptr = head_cnt;
      dst_addr[1] = addr_MSMC_fifo_PHY + offsetof(hlink_fifo_t, head_cnt);
      src_addr[1] = mca_btl_hlink_component.msmc_edma_transfer_addr + BTL_HLINK_MSMC_EDMA_MMAP_SIZE - sizeof(uint32_t);
      length[1]   = sizeof(uint32_t);
/***
      opal_output(0, "DMA_LINKED_COPY(%d):dst0 %08x src0 %08x len0 %d dst1 %08x src1 %08x len1 %d\n", head_cnt,
                      dst_addr[0], src_addr[0], length[0], dst_addr[1], src_addr[1], length[1]);
***/
#if 1
      btl_hlink_dma_copy (edma_hyplnk, dst_addr[0], src_addr[0], length[0]);
      btl_hlink_dma_copy (edma_hyplnk, dst_addr[1], src_addr[1], length[1]);
#else
      btl_hlink_dma_linked_copy (edma_hyplnk, 2, dst_addr, src_addr, length);
#endif
    } else {
      while(hlink_fifo_write(tmp_fifo, *tmp_rmt_msg_idx) != OMPI_SUCCESS)
      {  
        opal_output_verbose(20, mca_btl_base_output, "BTL_HLINK no room in FIFO (send)!");
      }
    }
    *tmp_rmt_msg_idx += btl_hlink_align64(sizeof(mca_btl_hlink_hdr_t) + seg_len_align);
    //Check for wrapparound
    if(*tmp_rmt_msg_idx > BTL_HLINK_PAYLOAD_SIZE) *tmp_rmt_msg_idx = 0;

    opal_output_verbose(20, mca_btl_base_output, "TIMPIDBG: hlink_module_send END hdr->len=%d hdr->tag=%d payload_size=%d!!!", hdr->len, hdr->tag, seg_len_align);
    mca_btl_hlink_frag_complete (frag);

    return 1;
}


/**
 * Initiate an inline send to the peer. If failure then return a descriptor.
 *
 * @param btl (IN)      BTL module
 * @param peer (IN)     BTL peer addressing
 */
int mca_btl_hlink_sendi( struct mca_btl_base_module_t* btl,
                      struct mca_btl_base_endpoint_t* endpoint,
                      struct opal_convertor_t* convertor,
                      void* header,
                      size_t header_size,
                      size_t payload_size,
                      uint8_t order,
                      uint32_t flags,
                      mca_btl_base_tag_t tag,
                      mca_btl_base_descriptor_t** descriptor )
{
    size_t length = (header_size + payload_size);
    mca_btl_hlink_frag_t* frag;
    int rc, rmt_ep;
    unsigned int seg_len_align;
    void *data_ptr = NULL;
    char *ptr;
    char host_name[80];
    unsigned int addr_RMT_PHY, addr_MSMC_fifo_PHY;
    int head_cnt;
    void *edma_hyplnk;
    void *tmp_fifo;
    int *tmp_rmt_msg_idx;


    opal_output_verbose(100, mca_btl_base_output, "TIMPIDBG: hlink_module_sendi A (header_size=%d payload_size=%d tag=%d)\n", 
       header_size, payload_size, tag);
    /* this check should be unnecessary... turn into an assertion? */
    assert (length < mca_btl_hlink_component.eager_limit);
    assert (0 == (flags & MCA_BTL_DES_SEND_ALWAYS_CALLBACK));
    //frag->base.des_flags = flags | MCA_BTL_DES_FLAGS_BTL_OWNERSHIP;   /* why do any flags matter here other than OWNERSHIP? */
 
    /* allocate a fragment, giving up if we can't get one */
    /* note that frag==NULL is equivalent to rc returning an error code */
    MCA_BTL_HLINK_FRAG_ALLOC_EAGER(frag, rc);
    if( OPAL_UNLIKELY(NULL == frag) ) {
        *descriptor = NULL;
        return rc;
    }

    /* fill in fragment fields */
    frag->segment.seg_len = length;
    frag->hdr->len = length;
    frag->hdr->tag = tag;

    rmt_ep = mca_btl_hlink_find_rmt_ep ((int)mca_btl_hlink_component.endpoint_count, (int)endpoint->hlink_iface, 
                                        (int)mca_btl_hlink_component.my_rank, (int)endpoint->peer_rank);

///
    if(endpoint->bridging == 1) {
      ptr = (char *)mca_btl_hlink_component.hl_win_rmt_wr_bridge_msgbase[rmt_ep] + mca_btl_hlink_component.hl_win_rmt_wr_bridge_msg_idx[rmt_ep];
      addr_RMT_PHY = mca_btl_hlink_component.ddr_rmt_bridge_addr + ((unsigned int)ptr - (unsigned int)mca_btl_hlink_component.hl_win_rmt_wr_bridge);
      tmp_fifo = mca_btl_hlink_component.fifo_bridge[1][rmt_ep];
      addr_MSMC_fifo_PHY = mca_btl_hlink_component.msmc_rmt_bridge_addr + ((unsigned int)tmp_fifo - (unsigned int)mca_btl_hlink_component.hl_win_msmc_rmt_wr_bridge);
      tmp_rmt_msg_idx = &mca_btl_hlink_component.hl_win_rmt_wr_bridge_msg_idx[rmt_ep];
      edma_hyplnk = endpoint->hlink_iface ? mca_btl_hlink_component.mpm_transport_hyplnk0 : mca_btl_hlink_component.mpm_transport_hyplnk1;
    } else {
      if(mca_btl_hlink_component.use_hl0 && (endpoint->hlink_iface == 0)) {
        ptr = (char *)mca_btl_hlink_component.hl_win_rmt_msgbase[rmt_ep] + mca_btl_hlink_component.hl_win_rmt_msg_idx[rmt_ep];
        addr_RMT_PHY = mca_btl_hlink_component.ddr_rmt_addr + ((unsigned int)ptr - (unsigned int)mca_btl_hlink_component.hl_win_rmt);
        tmp_fifo = mca_btl_hlink_component.fifo[1][rmt_ep];
        addr_MSMC_fifo_PHY = mca_btl_hlink_component.msmc_rmt_addr + ((unsigned int)tmp_fifo - (unsigned int)mca_btl_hlink_component.hl_win_msmc_rmt);
        tmp_rmt_msg_idx = &mca_btl_hlink_component.hl_win_rmt_msg_idx[rmt_ep];
        edma_hyplnk = mca_btl_hlink_component.mpm_transport_hyplnk0;
      } else if(mca_btl_hlink_component.use_hl1 && (endpoint->hlink_iface == 1)) {
        ptr = (char *)mca_btl_hlink_component.hl_win_rmt1_msgbase[rmt_ep] + mca_btl_hlink_component.hl_win_rmt1_msg_idx[rmt_ep];
        addr_RMT_PHY = mca_btl_hlink_component.ddr_rmt1_addr + ((unsigned int)ptr - (unsigned int)mca_btl_hlink_component.hl_win_rmt1);
        tmp_fifo = mca_btl_hlink_component.fifo1[1][rmt_ep];
        addr_MSMC_fifo_PHY = mca_btl_hlink_component.msmc_rmt1_addr + ((unsigned int)tmp_fifo - (unsigned int)mca_btl_hlink_component.hl_win_msmc_rmt1);
        tmp_rmt_msg_idx = &mca_btl_hlink_component.hl_win_rmt1_msg_idx[rmt_ep];
        edma_hyplnk = mca_btl_hlink_component.mpm_transport_hyplnk1;
      } else {
        gethostname(host_name, 80);
        opal_output(0, "\nBTL_HLINK(%s) endpoint invalid translation (hl0=%d hl1=%d hlink_iface=%d)\n", host_name,
          mca_btl_hlink_component.use_hl0, mca_btl_hlink_component.use_hl1, endpoint->hlink_iface); 
        return OMPI_ERR_OUT_OF_RESOURCE;
      }
    }
    opal_output_verbose(100, mca_btl_base_output, "TMPIDBG: sendi (hyplnk_iface%d) (bridging %d)", endpoint->hlink_iface, endpoint->bridging);
    if(mca_btl_hlink_component.use_edma && (frag->segment.seg_len >=  mca_btl_hlink_component.eager_limit))
    { // Big chunks of data first go to MSMC then to remote DDR over hyperlink using EDMA
      ptr = mca_btl_hlink_component.hl_win_edma_transfer; //user space address of MSMC transfer region
    }

    memcpy (ptr, frag->hdr, sizeof(mca_btl_hlink_hdr_t));
    ptr += sizeof(mca_btl_hlink_hdr_t);
    memcpy (ptr, header, header_size);
    ptr += header_size;

    /* bypassing the convertor may speed things up a little */
    if(payload_size) {
      opal_convertor_get_current_pointer (convertor, &data_ptr);
      seg_len_align = btl_hlink_align32 (payload_size);
      memcpy (ptr, data_ptr, seg_len_align);
    } else seg_len_align = 0;

    if(mca_btl_hlink_component.use_edma && (frag->segment.seg_len >=  mca_btl_hlink_component.eager_limit))
    { /** EDMA transfer **/
      unsigned int *head_ptr, dst_addr[2], src_addr[2], length[2];    
      opal_output_verbose(100, mca_btl_base_output, "\nBTL_HLINK(%s) sendi EDMA_TRANSFER (to %08x from %08x, size=%d!\n", host_name,
                  addr_RMT_PHY, mca_btl_hlink_component.msmc_edma_transfer_addr, 
                  btl_hlink_align64(sizeof(mca_btl_hlink_hdr_t) + header_size + seg_len_align));
      /* This is double linked transfer: push a fragment to remote DDR, then push a mail to remote MSMC */
      while((head_cnt = hlink_fifo_acquire(tmp_fifo, *tmp_rmt_msg_idx)) < 0)
      {
         opal_output(0, "BTL_HLINK no room in FIFO (send)!");
      }

      head_ptr = (unsigned int *)((char *)mca_btl_hlink_component.hl_win_edma_transfer + BTL_HLINK_MSMC_EDMA_MMAP_SIZE - sizeof(uint32_t));
      dst_addr[0] = addr_RMT_PHY;
      src_addr[0] = mca_btl_hlink_component.msmc_edma_transfer_addr;
      length[0]   = btl_hlink_align64(sizeof(mca_btl_hlink_hdr_t) + header_size + seg_len_align);
      head_cnt    ++;
      *head_ptr = head_cnt;
      dst_addr[1] = addr_MSMC_fifo_PHY + offsetof(hlink_fifo_t, head_cnt);
      src_addr[1] = mca_btl_hlink_component.msmc_edma_transfer_addr + BTL_HLINK_MSMC_EDMA_MMAP_SIZE - sizeof(uint32_t);
      length[1]   = sizeof(uint32_t);
/***
      opal_output(0, "DMA_LINKED_COPYi(%d):dst0 %08x src0 %08x len0 %d dst1 %08x src1 %08x len1 %d\n", head_cnt,
                      dst_addr[0], src_addr[0], length[0], dst_addr[1], src_addr[1], length[1]);
***/
#if 1
      btl_hlink_dma_copy (edma_hyplnk, dst_addr[0], src_addr[0], length[0]);
      btl_hlink_dma_copy (edma_hyplnk, dst_addr[1], src_addr[1], length[1]);
#else
      btl_hlink_dma_linked_copy (edma_hyplnk, 2, dst_addr, src_addr, length);
#endif

      //This is double-linked transfer, first moving a fragment and then pushing a main
      //btl_hlink_dma_copy (edma_hyplnk, addr_RMT_PHY, mca_btl_hlink_component.msmc_edma_transfer_addr, 
      //                    btl_hlink_align64(sizeof(mca_btl_hlink_hdr_t) + header_size + seg_len_align));
    } else {
      while(hlink_fifo_write(tmp_fifo, *tmp_rmt_msg_idx) != OMPI_SUCCESS)
      {
         gethostname(host_name, 80);
         opal_output(0, "\nBTL_HLINK(%s) no room in FIFO (send)!\n", host_name);
      }
    }
    *tmp_rmt_msg_idx += btl_hlink_align64(sizeof(mca_btl_hlink_hdr_t) + header_size + seg_len_align);
    //Check for wrapparound
    if(*tmp_rmt_msg_idx > BTL_HLINK_PAYLOAD_SIZE) *tmp_rmt_msg_idx = 0;

   opal_output_verbose(100, mca_btl_base_output, "TIMPIDBG: hlink_module_sendi C (header_size=%d payload_size=%d tag=%d)",
               header_size, payload_size, tag);
    mca_btl_hlink_frag_complete (frag);

    return OMPI_SUCCESS;
}

/******************************************************************************/
int mca_btl_hlink_ft_event(int state) {
    if(OPAL_CRS_CHECKPOINT == state) {
        ;
    }
    else if(OPAL_CRS_CONTINUE == state) {
        ;
    }
    else if(OPAL_CRS_RESTART == state) {
        ;
    }
    else if(OPAL_CRS_TERM == state ) {
        ;
    }
    else {
        ;
    }
    opal_output_verbose(100, mca_btl_base_output, "TIMPIDBG: hlink_module_ft_event???!!!");
    return OMPI_SUCCESS;
}
/*** nothing past this point ***/
