# -*- shell-script -*-
#
# Copyright (c) 2009      The University of Tennessee and The University
#                         of Tennessee Research Foundation.  All rights
#                         reserved.
# Copyright (c) 2009-2010 Cisco Systems, Inc.  All rights reserved.
# Copyright (c) 2011      Los Alamos National Security, LLC.
#                         All rights reserved.
# $COPYRIGHT$
#
# Additional copyrights may follow
#
# $HEADER$
#

# OMPI_CHECK_MPMTRANSPORT(prefix, [action-if-found], [action-if-not-found])
# --------------------------------------------------------
# check if MPMTRANSPORT support can be found.  sets prefix_{CPPFLAGS,
# LDFLAGS, LIBS} as needed and runs action-if-found if there is
# support, otherwise executes action-if-not-found
#   OMPI_CHECK_PACKAGE([$1],[mpm_transport.h],[mpmtransport],[mpm_transport_open],[-l hyplnk_k2h]
AC_DEFUN([OMPI_CHECK_MPMTRANSPORT], [
#    OPAL_VAR_SCOPE_PUSH([ompi_check_mpmtransport_happy ompi_check_mpmtransport_$1_save_CPPFLAGS ompi_check_mpmtransport_dir])

    AC_ARG_WITH([mpmtransport],
                [AC_HELP_STRING([--with-mpmtransport(=DIR)],
                [Build with MPMTRANSPORT kernel module support, searching for headers in DIR])])
    OMPI_CHECK_WITHDIR([mpmtransport], [$with_mpmtransport], [include/mpm_transport.h])

    AC_ARG_WITH([mpmtransport-libdir],
                [AC_HELP_STRING([--with-mpmtransport-libdir=DIR],
                                [Search for MPMTRANSPORT library in DIR])])
    OMPI_CHECK_WITHDIR([mpmtransport-libdir], [$with_mpmtransport_libdir], [libmpmtransport.*])

    AS_IF([test "$with_mpmtransport" != "no"],
          [AS_IF([test ! -z "$with_mpmtransport" -a "$with_mpmtransport" != "yes"],
                 [ompi_check_mpmtransport_dir="$with_mpmtransport"])
	   AS_IF([test ! -z "$with_mpmtransport_libdir" -a "$with_mpmtransport_libdir" != "yes"],
	         [ompi_check_mpmtransport_libdir="$with_mpmtransport_libdir"])
	   OMPI_CHECK_PACKAGE([$1],[mpm_transport.h],[mpmtransport],[mpm_transport_open],[],
                              [$ompi_check_mpmtransport_dir],[$ompi_check_mpmtransport_libdir],
			      [ompi_check_mpmtransport_happy="yes"], [ompi_check_mpmtransport_happy="no"])],
          [ompi_check_mpmtransport_happy="no"])

    AS_IF([test "$ompi_check_mpmtransport_happy" = "yes"],
          [$2],
          [AS_IF([test ! -z "$with_mpmtransport" -a "$with_mpmtransport" != "no"],
                 [AC_MSG_ERROR([MPMTRANSPORT support requested but not found.  Aborting])])
           $3])
#    OPAL_VAR_SCOPE_POP
])dnl

# MCA_btl_sm_CONFIG([action-if-can-compile],
#                   [action-if-cant-compile])
# ------------------------------------------------
AC_DEFUN([MCA_ompi_btl_hlink_CONFIG],[
    AC_CONFIG_FILES([ompi/mca/btl/hlink/Makefile])

    OPAL_VAR_SCOPE_PUSH([btl_hlink_mpmtransport_happy])
    OMPI_CHECK_MPMTRANSPORT([btl_hlink],
                     [btl_hlink_mpmtransport_happy=1],
                     [btl_hlink_mpmtransport_happy=0])

    AC_DEFINE_UNQUOTED([OMPI_BTL_HLINK_HAVE_MPMTRANSPORT],
                       [$btl_hlink_mpmtransport_happy],
                       [If MPMTRANSPORT support can be enabled within hlink])

    # at this point, we can only build hlink if we have MPMTRANSPORT support
    AS_IF([test "$btl_hlink_mpmtransport_happy" = "1"],
          [btl_hlink_WRAPPER_EXTRA_LDFLAGS="$btl_hlink_LDFLAGS"
           btl_hlink_WRAPPER_EXTRA_LIBS="$btl_hlink_LIBS"
           $1],
          [$2])

    # substitute in the things needed to build with MPMTRANSPORT support
    AC_SUBST([btl_hlink_CFLAGS])
    AC_SUBST([btl_hlink_CPPFLAGS])
    AC_SUBST([btl_hlink_LDFLAGS])
    AC_SUBST([btl_hlink_LIBS])

#    OPAL_VAR_SCOPE_POP
])dnl
