/*
 * Copyright (c) 2004-2005 The Trustees of Indiana University and Indiana
 *                         University Research and Technology
 *                         Corporation.  All rights reserved.
 * Copyright (c) 2004-2008 The University of Tennessee and The University
 *                         of Tennessee Research Foundation.  All rights
 *                         reserved.
 * Copyright (c) 2004-2005 High Performance Computing Center Stuttgart, 
 *                         University of Stuttgart.  All rights reserved.
 * Copyright (c) 2004-2005 The Regents of the University of California.
 *                         All rights reserved.
 * $COPYRIGHT$
 * 
 * Additional copyrights may follow
 * 
 * $HEADER$
 */
/**
 * @file
 */
#ifndef MCA_BTL_SRIO_SEND_FRAG_H
#define MCA_BTL_SRIO_SEND_FRAG_H

#include <sys/types.h>
#include "ompi/class/ompi_free_list.h"
#include "btl_srio.h"

#define MCA_BTL_SRIO_FRAG_TYPE_MASK ((uintptr_t)0x3)
#define MCA_BTL_SRIO_FRAG_SEND ((uintptr_t)0x0)
#define MCA_BTL_SRIO_FRAG_ACK ((uintptr_t)0x1)
#define MCA_BTL_SRIO_FRAG_PUT ((uintptr_t)0x2)
#define MCA_BTL_SRIO_FRAG_GET ((uintptr_t)0x3)

#define MCA_BTL_SRIO_FRAG_STATUS_MASK ((uintptr_t)0x4)


struct mca_btl_srio_frag_t;

struct mca_btl_srio_hdr_t {
    struct mca_btl_srio_frag_t *frag;
    size_t len;
    mca_btl_base_tag_t tag;
};
typedef struct mca_btl_srio_hdr_t mca_btl_srio_hdr_t;

/**
 * shared memory send fragment derived type.
 */
struct mca_btl_srio_frag_t {
    mca_btl_base_descriptor_t base;
    mca_btl_base_segment_t segment;
    struct mca_btl_base_endpoint_t *endpoint;
    size_t size;
    /* pointer written to the FIFO, this is the base of the shared memory region */
    mca_btl_srio_hdr_t *hdr;
    ompi_free_list_t* my_list;
};


typedef struct mca_btl_srio_frag_t mca_btl_srio_frag_t;
typedef struct mca_btl_srio_frag_t mca_btl_srio_frag1_t;
typedef struct mca_btl_srio_frag_t mca_btl_srio_frag2_t;


OBJ_CLASS_DECLARATION(mca_btl_srio_frag_t);
OBJ_CLASS_DECLARATION(mca_btl_srio_frag1_t);
OBJ_CLASS_DECLARATION(mca_btl_srio_frag2_t);

void mca_btl_srio_frag_init (mca_btl_srio_frag_t *frag, mca_btl_srio_t *srio_module);

#define MCA_BTL_SRIO_FRAG_ALLOC_EAGER(frag, rc)                        \
{                                                                       \
    ompi_free_list_item_t* item;                                        \
    OMPI_FREE_LIST_GET(&mca_btl_srio_component.srio_frags_eager, item, rc); \
    frag = (mca_btl_srio_frag_t*)item;                                 \ 
}

#define MCA_BTL_SRIO_FRAG_ALLOC_MAX(frag, rc)                          \
{                                                                       \
    ompi_free_list_item_t* item;                                        \
    OMPI_FREE_LIST_GET(&mca_btl_srio_component.srio_frags_max, item, rc);   \
    frag = (mca_btl_srio_frag_t*)item;                                    \
}

#define MCA_BTL_SRIO_FRAG_RETURN(frag)                                 \
{                                                                       \
    OMPI_FREE_LIST_RETURN(frag->my_list, (ompi_free_list_item_t*)(frag)); \
}


#define MCA_BTL_SRIO_FRAG_RETURN_EAGER(frag)                                 \
{                                                                            \
    OMPI_FREE_LIST_RETURN(&mca_btl_srio_component.srio_frags_eager,          \
                          (ompi_free_list_item_t*)(frag));                   \
    frag->segment.seg_addr.pval = frag+1;                                    \
}


static inline void mca_btl_srio_frag_complete (mca_btl_srio_frag_t *frag) {
    if (OPAL_UNLIKELY(MCA_BTL_DES_SEND_ALWAYS_CALLBACK & frag->base.des_flags)) {
        /* completion callback */
        frag->base.des_cbfunc(&mca_btl_srio.super, frag->endpoint,
                              &frag->base, OMPI_SUCCESS);
    }

    if (OPAL_LIKELY(frag->base.des_flags & MCA_BTL_DES_FLAGS_BTL_OWNERSHIP)) {
        MCA_BTL_SRIO_FRAG_RETURN(frag);
    }
}

#endif
